#!/usr/bin/perl

###############################################################################
#                                                                             #
# CR Tool: Partial Coreference Resolution Tool with NER Correction Heuristics #
#                                                                             #
###############################################################################

# Insert the full path for working from another folder
my $path = "./";

# Required files
require "$path/lib/GenericFunctions.pm";       # Required functions for extracting heads, merging entities, etc.
require "$path/lib/LinguisticFunctions.pm";    # Required functions for extracting heads, merging entities, etc.
require "$path/lib/MentionIdentification_f.pm";# Function for identifying mentions and extracting their main features (basic)
require "$path/lib/MentionExtract.pm";	       # Function for extracting the main features of the mentions

#-----------------------------------------------
# Options:
# Language (required):
#   Spanish: -s
#   Portuguese: -p
#   Galician: -g
#   English: -e
# Input:
#   CoNLL: 3 columns (default): -f
#   CoNLL: 11 columns (not implemented in this version): -F
# Output:
#   Coreference (default): -c
#   NER Correction: -n
use Getopt::Std;
our($opt_s, $opt_e, $opt_p, $opt_g, $opt_f, $opt_c, $opt_n);
$Getopt::Std::STANDARD_HELP_VERSION = 1;
getopts('epgsfcn');

sub HELP_MESSAGE {
    my $fh = shift;
    print $fh "Usage: cat corpus.txt | ./crtool.pl [options] > output.txt\n\n";
    print $fh "The input is a corpus with one token per line where each document\n";
    print $fh "starts with \"#begin document ID\" and ends with \"#end document ID\"\n";
    print $fh "Each token line should start with the number of the token position in\n";
    print $fh "the sentence, followed by the token, lemma and NER label separated by spaces.\n\n";
    print $fh "\tYou can test the system with the provided test file as follows:\n";
    print $fh "\t cat test.txt | ./crtool.pl -sf\n\n";
    print $fh "\tOptions:\n\n";
    print $fh "\tInput:\n";
    print $fh "\t-f \tFreeLing/CoNLL with three columns (default)\n\n";
    #print $fh "\t-F \tFull, with eleven columns\n\n";
    print $fh "\tOutput:\n";
    print $fh "\t-c \tCoreference and NER correction (default)\n";
    print $fh "\t-n \tOnly NER correction\n\n";
    print $fh "\tLanguage (required):\n";
    print $fh "\t-s \tSpanish\n";
    print $fh "\t-p \tPortuguese\n";
    print $fh "\t-g \tGalician\n";
    print $fh "\t-e \tEnglish\n\n";
}

sub VERSION_MESSAGE {
    my $fh = shift;
    print $fh "\nCR Tool 0.3\n";
}

# Invalid options > prints the help
if ( (!$opt_e && !$opt_p && !$opt_g && !$opt_s) ||
     ($opt_e && ($opt_p || $opt_g || $opt_s)) ||
     ($opt_p && ($opt_e || $opt_g || $opt_s)) ||
     ($opt_g && ($opt_e || $opt_p || $opt_s)) ||
     ($opt_c && $opt_n) ) {
    HELP_MESSAGE(*STDOUT);
    exit(1);
}
#-----------------------------------------------

# Default input (FreeLing)
if (!$opt_f && !$opt_F) {
    $opt_f = 1;
}
# Deafult output (Coreference)
if (!$opt_c && !$opt_n) {
    $opt_c = 1;
}
#

# Global variables
#-----------------
# Mentions
our (%Mentions, %MentionsForm, %MentionsPos, %PosMentions, %MentionsSent, %MentionsTag, %IDs, %Gender);
our (%Lema, %Tag, %Head, %HeadNP, %FullNP, %NP, %MentionsPosE, $Pos_start, $Pos_end);
# Entities
our (%FullNPEnt, %HeadNPEnt, %GenderEnt, %Size, %TagEnt);
# Misc
our ($GenOK, $CompNP, $CompNP_toks, $DistanceSent);
our ($NP_i, $NP_m, $FullNP_m, $FullNP_i, @FullNP_m, $CompFN, $Constraint, $num_sentence, $num_mention);
# Others
our ($separador, $artigo, $artigo_raw, $art_num, @Lines, $cont_linhas, $id_entity);
my %Linhas = ();
# Nouns
our ($PerNouns, $FemNames, $MaleNames);

## Loads the nouns (for mention identication)
if ($opt_e) {
    open(FILE, "$path/lex/person_en.txt") or die ("Cannot open the person names file\n");
} elsif ($opt_g) {
    open(FILE, "$path/lex/person_gl.txt") or die ("Cannot open the person names file\n");
} elsif ($opt_s) {
    open(FILE, "$path/lex/person_es.txt") or die ("Cannot open the person names file\n");
} elsif ($opt_p) {
    open(FILE, "$path/lex/person_pt.txt") or die ("Cannot open the person names file\n");
}
while ($line = <FILE>) {
    chomp($line);
    my $cont_per++;
    $PerNouns{$line} = $cont_per;
}
close(FILE);

## Loads the male/female nouns (now all of them)
open(FILE, "$path/lex/female_all_uniq.txt") or die ("Cannot open the female names file\n");
while ($line = <FILE>) {
    chomp($line);
    my $cont_fem++;
    $FemNames{$line} = $cont_fem;
}
close(FILE);
#
open(FILE, "$path/lex/male_all_uniq.txt") or die ("Cannot open the male names file\n");
while ($line = <FILE>) {
    chomp($line);
    my $cont_male++;
    $MaleNames{$line} = $cont_male;
}
close(FILE);
#####

# Articles must start with "#begin document"
# and end with "#end document" lines
# (you can run "prepare.pl" before)
$separador = /^#end document.*$/;
$/ = $separador;

#####################################################

# Each article is loaded (and counted) in $artigo_raw
while ($artigo_raw = <STDIN>) {
    $art_num++;
    chomp($artigo_raw);

    # The content of the article is extracted and goes to a hash ($artigo > %Linhas)
    ($artigo) = ($artigo_raw =~ /(#begin document.*\n[^#]+\n#end document.*)/);
    (@Lines) = split("\n", $artigo);

    if ($#Lines < 4) {
	print STDERR "WARNING: Possible wrong input\n";
    }
    
    %Linhas = ();
    foreach my $l(@Lines) {
	$cont_linhas++;
	$Linhas{$cont_linhas} = $l;
    }

    # First Step: Mention Identification
    #-----------------------------------
    # %Mentions{mention} = X (no entity)
    # %MentionsPos{mention} = position
    # %MentionsForm{mention} = token(s)
    ####################################
    if ($opt_f) {
	%Linhas = MentionIdentify_f(%Linhas);
    } # elsif ($opt_F) { MentionIdentify_F(); }
    %Linhas = MentionExtract(%Linhas);
    
    # Check: if the previous modules have not found
    # any named entity in the text: finish.
    my $size_mentions = keys %Mentions;
    if ($size_mentions < 2) {
	%Mentions = ();
    }

    # Coreference Modules
    #####################
    
    # Nominal Coreference Modules
    #############################
    
    # StringMatch
    #############
    # Starts in second mention and looks backwards for candidates
    foreach my $m(sort {$a <=> $b} keys %Mentions) {
	if ($m ne "1") {
	    my $i = $m - 1;
	    while ($i >= 1) {
		if (($MentionsPos{$m} != $MentionsPos{$i}) && ($m != $i)) {
		    if ($MentionsForm{$m} eq $MentionsForm{$i}) {
			#if ($Tag{$i} eq $Tag{$m}) {
			my $new_tag = SelectTag($m, $i);
			$Mentions = Merge(\%Mentions, $m, $i);
			%Mentions = %$Mentions;
			$TagEnt{$Mentions{$m}} = $new_tag;
			$i = 0;
			#}
		    }
		}
		$i--;
	    }
	}
    }
    # Proper Noun inclusion (longest proper noun of A
    # should be in longest proper noun of B -or vice-versa)
    #------------------------------------------------------
    
    # When selecting a mention (m), it should be
    # a singleton or the first mention of its entity
    foreach my $m(sort {$a <=> $b} keys %Mentions) {
	if ( ($m ne "1") &&
	     ( ($Mentions{$m} ne "X" && $Mentions{$m}[0] == $m) || ($Mentions{$m} eq "X") ) ) {
	    my $i = $m - 1;
	    while ($i >= 1) {
		if (defined $Mentions{$i}) {
		    #if ($Tag{$i} eq $Tag{$m}) {
		    ($CompNP, $CompNPToks) = CompareNP($m, $i); 
		    if ($CompNP == 1) {
			if (CompareGN($m, $i) == 1 && CompareTags($m, $i) == 1) {
			    my $new_tag = SelectTag($m, $i);
			    $Mentions = Merge(\%Mentions, $m, $i);
			    %Mentions = %$Mentions;
			    $TagEnt{$Mentions{$m}} = $new_tag;
			    last;
			}
		    }
		    #}
		}
		$i--;
	    }
	}
    }
    # End of the nominal modules (not baselines)
    
    # Default rules
    ###############
    
    # Singletons
    #-----------
    foreach my $m(sort {$a <=> $b} keys %Mentions) {
	if ($Mentions{$m} =~ /^X$/) {
	    # Default feats
	    $id_entity++;
	    $Mentions{$m} = $id_entity;
	    $TagEnt{$Mentions{$m}} = $Tag{$m};
	    $FullNPEnt{$Mentions{$m}}[0] = $MentionsForm{$m};
	}
    }
    
    $id_entity = 0;
    $num_sentence = 0;
    $num_mention = 0;
    
    # Get the longest NP of each entity
    ###################################
    foreach my $m(sort {$a <=> $b} keys %TagEnt) {
	my $size = ();
	my $LongNP_m = ();
	foreach my $f(@{$FullNPEnt{$m}}) {
	    if ($f =~ /_/) {
		my (@size) = split("_", $f);
		$size{$f} = $#size + 1;
		my $first = 0;
		foreach my $n(sort {$size{$b} <=> $size{$a} } keys %size ) {
		    if (!$first) {
			$LongNP_m = $n;
			$LongNPsize_m = $size{$n};
		    }
		    $first = 1;
		}
		@size = ();
		%size = ();
	    }
	}
	if (!$LongNP_m) {
	    $LongNP_m = $FullNPEnt{$m}[0];
	}
	$FullNPLong{$m} = $LongNP_m;
    }
    
    # Hash modification for printing
    #-------------------------------    
    foreach my $m(sort {$a <=> $b} keys %Mentions) {
	my $id_e = $Mentions{$m};
	my $final_tag = $TagEnt{$id_e};
	my $final_lemma = $FullNPLong{$Mentions{$m}};	
	my ($ps, $tk, $lm, $tg, $cr) = split(" ", $Linhas{$MentionsPos{$m}});
	# Coreference output
	if ($opt_c) {
	    $Linhas{$MentionsPos{$m}} = $ps . " " . $tk . " " . $lm . " " . $final_tag . " ($id_e)";
	}
	# NER output
	elsif ($opt_n) {
	    $Linhas{$MentionsPos{$m}} = $ps . " " . $tk . " " . $lm . " " . $final_tag;
	}
    }
    
    # Hash printing
    #--------------
    foreach $l(sort {$a <=> $b} keys %Linhas) {
	print "$Linhas{$l}\n";
    }
    print "\n";
    
    ####################################
    
    # Cleans the entity arrays
    foreach $e(values %Mentions) {
	@$e = ();
    }
    
    %Linhas = ();
    $cont_linhas = 0;
    
    # Mentions
    %Mentions = ();
    %MentionsPos = ();
    %MentionsTag = ();
    %PosMentions = ();
    %MentionsForm = ();
    %Gender = ();
    %HeadNP = ();
    %FullNP = ();
    %Tag = ();
    %TagEnt = ();
    %IDs = ();		
    
    # Entities
    %GenderEnt = ();
    %FullNPEnt = ();
    %HeadNPEnt = ();
}

####################################
